#!/s/std/bin/perl -w

# <std-header style='perl' orig-src='shore'>
#
#  $Id: GenSrcDirInfo.pl,v 1.8 1999/06/07 19:09:12 kupsch Exp $
#
# SHORE -- Scalable Heterogeneous Object REpository
#
# Copyright (c) 1994-99 Computer Sciences Department, University of
#                       Wisconsin -- Madison
# All Rights Reserved.
#
# Permission to use, copy, modify and distribute this software and its
# documentation is hereby granted, provided that both the copyright
# notice and this permission notice appear in all copies of the
# software, derivative works or modified versions, and any portions
# thereof, and that both notices appear in supporting documentation.
#
# THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
# OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
# "AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
# FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
#
# This software was developed with support by the Advanced Research
# Project Agency, ARPA order number 018 (formerly 8230), monitored by
# the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.
# Further funding for this work was provided by DARPA through
# Rome Research Laboratory Contract No. F30602-97-2-0247.
#
#   -- do not edit anything above this line --   </std-header>


use strict;
use Getopt::Long;

sub Usage
{
    my $progname = $0;
    $progname =~ s/.*[\\\/]//;
    print STDERR <<EOF;
Usage: $progname [--input=infilename] [--output=outfilename] [--createDirs] [--help]
Generates set of make variables from list of names and directories.

    --input         input file
    --ouput         output file
    --createDirs    creates dir for output file if it doesn't exist
    --help|h        prints this message and exits.
EOF
}

my %options = (input => '-', output => '-', createDirs => 0, help => 0);
my @options = ("input=s", "output=s", "createDirs!", "help|h!");
my $ok = GetOptions(\%options, @options);
$ok = 0 if $#ARGV != -1;
if (!$ok || $options{help})  {
    Usage();
    exit($ok == 0);
}


sub CreateDirs
{
    my $dir = shift;

    return if -e $dir;

    my $parentDir;
    ($parentDir) = ($dir =~ /(.*)\//);
    $parentDir = '.' if !$parentDir;

    CreateDirs($parentDir) unless -e $parentDir;

    my $mode = (stat $parentDir)[2] | 0700;
    mkdir $dir, $mode or die "mkdir $dir";
}


sub CanonicalFilename
{
    my $filename = shift;
    my @filename = split(/\//, $filename);
    foreach my $i (reverse 0 .. $#filename)  {
        splice(@filename, $i, 1) if $filename[$i] eq '.' || $i != 0 && $filename[$i] eq '';
    }
    my $i = 1;
    while ($i <= $#filename)  {
        if ($filename[$i] eq '..' && $filename[$i - 1] ne '..')  {
            splice(@filename, $i - 1, 2);
            --$i if $i > 1;
        }  else  {
            ++$i;
        }
    }
    $filename[0] = '.' if $#filename == -1;
    join('/', @filename);
}

my $input = $options{input};
my $output = $options{output};
my $input_dir = '-';
my $output_dir = '-';
$input_dir  = CanonicalFilename($input =~ /(.*)\//) if $input ne '-';
$output_dir = CanonicalFilename($output =~ /(.*)\//) if $output ne '-';
my $useBuildDir = ($input_dir ne $output_dir);
my $curTime = localtime;

CreateDirs($output =~ /(.*)\//) if defined $options{createDirs} && $output ne '-';


my $header = <<EOF;
/* DO NOT EDIT --- generated by $0 from $input
		   on $curTime

<std-header style='imake' orig-src='shore' genfile='true'>

SHORE -- Scalable Heterogeneous Object REpository

Copyright (c) 1994-99 Computer Sciences Department, University of
                      Wisconsin -- Madison
All Rights Reserved.

Permission to use, copy, modify and distribute this software and its
documentation is hereby granted, provided that both the copyright
notice and this permission notice appear in all copies of the
software, derivative works or modified versions, and any portions
thereof, and that both notices appear in supporting documentation.

THE AUTHORS AND THE COMPUTER SCIENCES DEPARTMENT OF THE UNIVERSITY
OF WISCONSIN - MADISON ALLOW FREE USE OF THIS SOFTWARE IN ITS
"AS IS" CONDITION, AND THEY DISCLAIM ANY LIABILITY OF ANY KIND
FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.

This software was developed with support by the Advanced Research
Project Agency, ARPA order number 018 (formerly 8230), monitored by
the U.S. Army Research Laboratory under contract DAAB07-91-C-Q518.
Further funding for this work was provided by DARPA through
Rome Research Laboratory Contract No. F30602-97-2-0247.

  -- do not edit anything above this line --   </std-header>*/

EOF

open INFILE, "<$input" or die "open $input";
open OUTFILE, ">$output" or die "open $output";

print OUTFILE $header;

while (<INFILE>)  {
    next if (/^\s*(#.*)?$/);

    if (/^\s*(\S*)\s*(.*)/)  {
	my ($name, $path) = ($1, $2);
	print OUTFILE <<EOF;

${name}_PATH           = $path
${name}_DIR            = \$(TOP)/\$(${name}_PATH)
${name}_BUILD_DIR      = \$(BUILD_TOP)/\$(${name}_PATH)
${name}_LIB            = \$(${name}_BUILD_DIR)/lib$name.LIBEXT
${name}_SHARED_LIB     = \$(${name}_BUILD_DIR)/lib$name.DLLEXT
${name}_INCL_DIR       = \$(${name}_DIR)
${name}_INCL_BUILD_DIR = \$(${name}_BUILD_DIR)
EOF

        print OUTFILE "${name}_INCL           = -I\$(${name}_INCL_DIR)";
	print OUTFILE " -I\$(${name}_INCL_BUILD_DIR)" if $useBuildDir;
	print OUTFILE "\n";
    }  else  {
	die;
    }
}

close OUTFILE or die "close $output";
close INFILE or die "close $input";
